[llvm] r231042 - Remap arguments and non-alloca values used by outlined C++ exception handlers.

Andrew Kaylor andrew.kaylor at intel.com
Mon Mar 2 16:41:04 PST 2015


Author: akaylor
Date: Mon Mar  2 18:41:03 2015
New Revision: 231042

URL: http://llvm.org/viewvc/llvm-project?rev=231042&view=rev
Log:
Remap arguments and non-alloca values used by outlined C++ exception handlers.

Differential Revision: http://reviews.llvm.org/D7844


Added:
    llvm/trunk/test/CodeGen/X86/cppeh-inalloca.ll
    llvm/trunk/test/CodeGen/X86/cppeh-nonalloca-frame-values.ll
Modified:
    llvm/trunk/lib/CodeGen/WinEHPrepare.cpp

Modified: llvm/trunk/lib/CodeGen/WinEHPrepare.cpp
URL: http://llvm.org/viewvc/llvm-project/llvm/trunk/lib/CodeGen/WinEHPrepare.cpp?rev=231042&r1=231041&r2=231042&view=diff
==============================================================================
--- llvm/trunk/test/CodeGen/X86/cppeh-inalloca.ll (added)
+++ llvm/trunk/test/CodeGen/X86/cppeh-inalloca.ll Mon Mar  2 18:41:03 2015
@@ -0,0 +1,181 @@
+; RUN: opt -mtriple=i386-pc-windows-msvc -winehprepare -S -o - < %s | FileCheck %s
+
+; This test is built from the following code:
+; struct A {
+;   A(int a);
+;   A(const A &o);
+;   ~A();
+;   int a;
+; };
+;
+; void may_throw();
+;
+; int test(A a) {
+;   try {
+;     may_throw();
+;   }
+;   catch (int e) {
+;     return a.a + e;
+;   }
+;   return 0;
+; }
+;
+; The test was built for a 32-bit Windows target and then the reference to
+; the inalloca instruction was manually sunk into the landingpad.
+
+; ModuleID = 'cppeh-inalloca.cpp'
+target datalayout = "e-m:w-p:32:32-i64:64-f80:32-n8:16:32-S32"
+target triple = "i386-pc-windows-msvc"
+
+%rtti.TypeDescriptor2 = type { i8**, i8*, [3 x i8] }
+%struct.A = type { i32 }
+
+$"\01??_R0H at 8" = comdat any
+
+@"\01??_7type_info@@6B@" = external constant i8*
+@"\01??_R0H at 8" = linkonce_odr global %rtti.TypeDescriptor2 { i8** @"\01??_7type_info@@6B@", i8* null, [3 x i8] c".H\00" }, comdat
+
+; The function entry should be rewritten like this.
+; CHECK: define i32 @"\01?test@@YAHUA@@@Z"(<{ %struct.A }>* inalloca) #0 {
+; CHECK: entry:
+; CHECK:   %frame.alloc = call i8* @llvm.frameallocate(i32 24)
+; CHECK:   %eh.data = bitcast i8* %frame.alloc to %"struct.\01?test@@YAHUA@@@Z.ehdata"*
+; CHECK:   %.tmp.reg2mem = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 3
+; CHECK:   %.tmp = select i1 true, <{ %struct.A }>* %0, <{ %struct.A }>* undef
+; CHECK:   store <{ %struct.A }>* %.tmp, <{ %struct.A }>** %.tmp.reg2mem
+; CHECK-NOT:  %retval = alloca i32, align 4
+; CHECK:   %retval = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 4
+; CHECK:   %exn.slot = alloca i8*
+; CHECK:   %ehselector.slot = alloca i32
+; CHECK-NOT:  %e = alloca i32, align 4
+; CHECK:   %e = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 2
+; CHECK:   %cleanup.dest.slot = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 5
+; CHECK:   invoke void @"\01?may_throw@@YAXXZ"()
+; CHECK:           to label %invoke.cont unwind label %lpad
+
+define i32 @"\01?test@@YAHUA@@@Z"(<{ %struct.A }>* inalloca) #0 {
+entry:
+  %retval = alloca i32, align 4
+  %exn.slot = alloca i8*
+  %ehselector.slot = alloca i32
+  %e = alloca i32, align 4
+  %cleanup.dest.slot = alloca i32
+  invoke void @"\01?may_throw@@YAXXZ"()
+          to label %invoke.cont unwind label %lpad
+
+invoke.cont:                                      ; preds = %entry
+  br label %try.cont
+
+lpad:                                             ; preds = %entry
+  %1 = landingpad { i8*, i32 } personality i8* bitcast (i32 (...)* @__CxxFrameHandler3 to i8*)
+          cleanup
+          catch i8* bitcast (%rtti.TypeDescriptor2* @"\01??_R0H at 8" to i8*)
+  %2 = extractvalue { i8*, i32 } %1, 0
+  store i8* %2, i8** %exn.slot
+  %3 = extractvalue { i8*, i32 } %1, 1
+  store i32 %3, i32* %ehselector.slot
+  br label %catch.dispatch
+
+catch.dispatch:                                   ; preds = %lpad
+  %sel = load i32* %ehselector.slot
+  %4 = call i32 @llvm.eh.typeid.for(i8* bitcast (%rtti.TypeDescriptor2* @"\01??_R0H at 8" to i8*)) #3
+  %matches = icmp eq i32 %sel, %4
+  br i1 %matches, label %catch, label %ehcleanup
+
+catch:                                            ; preds = %catch.dispatch
+  %exn = load i8** %exn.slot
+  %5 = call i8* @llvm.eh.begincatch(i8* %exn) #3
+  %6 = bitcast i8* %5 to i32*
+  %7 = load i32* %6, align 4
+  store i32 %7, i32* %e, align 4
+  %a = getelementptr inbounds <{ %struct.A }>* %0, i32 0, i32 0
+  %a1 = getelementptr inbounds %struct.A* %a, i32 0, i32 0
+  %8 = load i32* %a1, align 4
+  %9 = load i32* %e, align 4
+  %add = add nsw i32 %8, %9
+  store i32 %add, i32* %retval
+  store i32 1, i32* %cleanup.dest.slot
+  call void @llvm.eh.endcatch() #3
+  br label %cleanup
+
+try.cont:                                         ; preds = %invoke.cont
+  store i32 0, i32* %retval
+  store i32 1, i32* %cleanup.dest.slot
+  br label %cleanup
+
+; The cleanup block should be re-written like this.
+; CHECK: cleanup:                                          ; preds = %try.cont, %catch
+; CHECK-NOT:  %a2 = getelementptr inbounds <{ %struct.A }>* %0, i32 0, i32 0
+; CHECK:   %.tmp.reload1 = load volatile <{ %struct.A }>** %.tmp.reg2mem
+; CHECK:   %a2 = getelementptr inbounds <{ %struct.A }>* %.tmp.reload1, i32 0, i32 0
+; CHECK:   call x86_thiscallcc void @"\01??1A@@QAE at XZ"(%struct.A* %a2) #2
+; CHECK:   %10 = load i32* %retval
+; CHECK:   ret i32 %10
+
+cleanup:                                          ; preds = %try.cont, %catch
+  %a2 = getelementptr inbounds <{ %struct.A }>* %0, i32 0, i32 0
+  call x86_thiscallcc void @"\01??1A@@QAE at XZ"(%struct.A* %a2) #3
+  %10 = load i32* %retval
+  ret i32 %10
+
+ehcleanup:                                        ; preds = %catch.dispatch
+  %a3 = getelementptr inbounds <{ %struct.A }>* %0, i32 0, i32 0
+  call x86_thiscallcc void @"\01??1A@@QAE at XZ"(%struct.A* %a3) #3
+  br label %eh.resume
+
+eh.resume:                                        ; preds = %ehcleanup
+  %exn2 = load i8** %exn.slot
+  %sel3 = load i32* %ehselector.slot
+  %lpad.val = insertvalue { i8*, i32 } undef, i8* %exn2, 0
+  %lpad.val4 = insertvalue { i8*, i32 } %lpad.val, i32 %sel3, 1
+  resume { i8*, i32 } %lpad.val4
+}
+
+; The following catch handler should be outlined.
+; CHECK: define i8* @"\01?test@@YAHUA@@@Z.catch"(i8*, i8*) {
+; CHECK: catch.entry:
+; CHECK:   %eh.alloc = call i8* @llvm.framerecover(i8* bitcast (i32 (<{ %struct.A }>*)* @"\01?test@@YAHUA@@@Z" to i8*), i8* %1)
+; CHECK:   %eh.data = bitcast i8* %eh.alloc to %"struct.\01?test@@YAHUA@@@Z.ehdata"*
+; CHECK:   %eh.obj.ptr = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 1
+; CHECK:   %eh.obj = load i8** %eh.obj.ptr
+; CHECK:   %e = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 2
+; CHECK:   %eh.temp.alloca = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 3
+; CHECK:   %.reload = load <{ %struct.A }>** %eh.temp.alloca
+; CHECK:   %retval = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 4
+; CHECK:   %cleanup.dest.slot = getelementptr inbounds %"struct.\01?test@@YAHUA@@@Z.ehdata"* %eh.data, i32 0, i32 5
+; CHECK:   %2 = bitcast i8* %eh.obj to i32*
+; CHECK:   %3 = load i32* %2, align 4
+; CHECK:   store i32 %3, i32* %e, align 4
+; CHECK:   %a = getelementptr inbounds <{ %struct.A }>* %.reload, i32 0, i32 0
+; CHECK:   %a1 = getelementptr inbounds %struct.A* %a, i32 0, i32 0
+; CHECK:   %4 = load i32* %a1, align 4
+; CHECK:   %5 = load i32* %e, align 4
+; CHECK:   %add = add nsw i32 %4, %5
+; CHECK:   store i32 %add, i32* %retval
+; CHECK:   store i32 1, i32* %cleanup.dest.slot
+; CHECK:   ret i8* blockaddress(@"\01?test@@YAHUA@@@Z", %cleanup)
+; CHECK: }
+
+
+declare void @"\01?may_throw@@YAXXZ"() #0
+
+declare i32 @__CxxFrameHandler3(...)
+
+; Function Attrs: nounwind readnone
+declare i32 @llvm.eh.typeid.for(i8*) #1
+
+declare i8* @llvm.eh.begincatch(i8*)
+
+declare void @llvm.eh.endcatch()
+
+; Function Attrs: nounwind
+declare x86_thiscallcc void @"\01??1A@@QAE at XZ"(%struct.A*) #2
+
+attributes #0 = { "less-precise-fpmad"="false" "no-frame-pointer-elim"="true" "no-frame-pointer-elim-non-leaf" "no-infs-fp-math"="false" "no-nans-fp-math"="false" "stack-protector-buffer-size"="8" "unsafe-fp-math"="false" "use-soft-float"="false" }
+attributes #1 = { nounwind readnone }
+attributes #2 = { nounwind "less-precise-fpmad"="false" "no-frame-pointer-elim"="true" "no-frame-pointer-elim-non-leaf" "no-infs-fp-math"="false" "no-nans-fp-math"="false" "stack-protector-buffer-size"="8" "unsafe-fp-math"="false" "use-soft-float"="false" }
+attributes #3 = { nounwind }
+
+!llvm.ident = !{!0}
+
+!0 = !{!"clang version 3.7.0 (trunk 228868)"}

Added: llvm/trunk/test/CodeGen/X86/cppeh-nonalloca-frame-values.ll
URL: http://llvm.org/viewvc/llvm-project/llvm/trunk/test/CodeGen/X86/cppeh-nonalloca-frame-values.ll?rev=231042&view=auto
==============================================================================
--- llvm/trunk/test/CodeGen/X86/cppeh-nonalloca-frame-values.ll (added)
+++ llvm/trunk/test/CodeGen/X86/cppeh-nonalloca-frame-values.ll Mon Mar  2 18:41:03 2015
@@ -0,0 +1,264 @@
+; RUN: opt -mtriple=x86_64-pc-windows-msvc -winehprepare -S -o - < %s | FileCheck %s
+
+; This test is based on the following code:
+;
+; struct SomeData {
+;   int a;
+;   int b;
+; };
+; 
+; void may_throw();
+; void does_not_throw(int i);
+; void dump(int *, int, SomeData&);
+; 
+; void test() {
+;   int NumExceptions = 0;
+;   int ExceptionVal[10];
+;   SomeData Data = { 0, 0 };
+; 
+;   for (int i = 0; i < 10; ++i) {
+;     try {
+;       may_throw();
+;       Data.a += i;
+;     }
+;     catch (int e) {
+;       ExceptionVal[NumExceptions] = e;
+;       ++NumExceptions;
+;       if (e == i)
+;         Data.b += e;
+;       else
+;         Data.a += e;
+;     }
+;     does_not_throw(NumExceptions);
+;   }
+;   dump(ExceptionVal, NumExceptions, Data);
+; }
+;
+; Unlike the cppeh-frame-vars.ll test, this test was generated using -O2
+; optimization, which results in non-alloca values being used in the
+; catch handler.
+
+; ModuleID = 'cppeh-frame-vars.cpp'
+target datalayout = "e-m:e-i64:64-f80:128-n8:16:32:64-S128"
+target triple = "x86_64-pc-windows-msvc"
+
+%rtti.TypeDescriptor2 = type { i8**, i8*, [3 x i8] }
+%struct.SomeData = type { i32, i32 }
+
+$"\01??_R0H at 8" = comdat any
+
+@"\01??_7type_info@@6B@" = external constant i8*
+@"\01??_R0H at 8" = linkonce_odr global %rtti.TypeDescriptor2 { i8** @"\01??_7type_info@@6B@", i8* null, [3 x i8] c".H\00" }, comdat
+
+; This structure should be declared for the frame allocation block.
+; CHECK: %"struct.\01?test@@YAXXZ.ehdata" = type { i32, i8*, i32, [10 x i32], i32, i32*, i32* }
+
+; The function entry should be rewritten like this.
+; CHECK: define void @"\01?test@@YAXXZ"() #0 {
+; CHECK: entry:
+; CHECK:  %frame.alloc = call i8* @llvm.frameallocate(i32 80)
+; CHECK:  %eh.data = bitcast i8* %frame.alloc to %"struct.\01?test@@YAXXZ.ehdata"*
+; CHECK-NOT:  %ExceptionVal = alloca [10 x i32], align 16
+; CHECK:  %NumExceptions.020.reg2mem = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 2
+; CHECK:  %i.019.reg2mem = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 4
+; CHECK:  %ExceptionVal = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 3
+; CHECK:  %Data = alloca i64, align 8
+; CHECK:  %tmpcast = bitcast i64* %Data to %struct.SomeData*
+; CHECK:  %0 = bitcast [10 x i32]* %ExceptionVal to i8*
+; CHECK:  call void @llvm.lifetime.start(i64 40, i8* %0) #1
+; CHECK:  store i64 0, i64* %Data, align 8
+; CHECK:  %a.reg2mem = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 5
+; CHECK:  %a = bitcast i64* %Data to i32*
+; CHECK:  store i32* %a, i32** %a.reg2mem
+; CHECK:  %b.reg2mem = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 6
+; CHECK:  %b = getelementptr inbounds %struct.SomeData* %tmpcast, i64 0, i32 1
+; CHECK:  store i32* %b, i32** %b.reg2mem
+; CHECK:  store i32 0, i32* %NumExceptions.020.reg2mem
+; CHECK:  store i32 0, i32* %i.019.reg2mem
+; CHECK:  br label %for.body
+
+; Function Attrs: uwtable
+define void @"\01?test@@YAXXZ"() #0 {
+entry:
+  %ExceptionVal = alloca [10 x i32], align 16
+  %Data = alloca i64, align 8
+  %tmpcast = bitcast i64* %Data to %struct.SomeData*
+  %0 = bitcast [10 x i32]* %ExceptionVal to i8*
+  call void @llvm.lifetime.start(i64 40, i8* %0) #1
+  store i64 0, i64* %Data, align 8
+  %a = bitcast i64* %Data to i32*
+  %b = getelementptr inbounds %struct.SomeData* %tmpcast, i64 0, i32 1
+  br label %for.body
+
+; CHECK: for.body:
+; CHECK-NOT:  %NumExceptions.020 = phi i32 [ 0, %entry ], [ %NumExceptions.1, %try.cont ]
+; CHECK-NOT:  %i.019 = phi i32 [ 0, %entry ], [ %inc5, %try.cont ]
+; CHECK:  %i.019.reload = load i32* %i.019.reg2mem
+; CHECK:  %NumExceptions.020.reload = load i32* %NumExceptions.020.reg2mem
+for.body:                                         ; preds = %entry, %try.cont
+  %NumExceptions.020 = phi i32 [ 0, %entry ], [ %NumExceptions.1, %try.cont ]
+  %i.019 = phi i32 [ 0, %entry ], [ %inc5, %try.cont ]
+  invoke void @"\01?may_throw@@YAXXZ"()
+          to label %invoke.cont unwind label %lpad
+
+; CHECK: invoke.cont:                                      ; preds = %for.body
+; CHECK-NOT:  %1 = load i32* %a, align 8, !tbaa !2
+; CHECK-NOT:  %add = add nsw i32 %1, %i.019
+; CHECK-NOT:  store i32 %add, i32* %a, align 8, !tbaa !2
+; CHECK:   %a.reload3 = load volatile i32** %a.reg2mem
+; CHECK:   %1 = load i32* %a.reload3, align 8, !tbaa !2
+; CHECK:   %add = add nsw i32 %1, %i.019.reload
+; CHECK:   %a.reload2 = load volatile i32** %a.reg2mem
+; CHECK:   store i32 %add, i32* %a.reload2, align 8, !tbaa !2
+; CHECK:   br label %try.cont
+invoke.cont:                                      ; preds = %for.body
+  %1 = load i32* %a, align 8, !tbaa !2
+  %add = add nsw i32 %1, %i.019
+  store i32 %add, i32* %a, align 8, !tbaa !2
+  br label %try.cont
+
+lpad:                                             ; preds = %for.body
+  %2 = landingpad { i8*, i32 } personality i8* bitcast (i32 (...)* @__CxxFrameHandler3 to i8*)
+          catch i8* bitcast (%rtti.TypeDescriptor2* @"\01??_R0H at 8" to i8*)
+  %3 = extractvalue { i8*, i32 } %2, 1
+  %4 = tail call i32 @llvm.eh.typeid.for(i8* bitcast (%rtti.TypeDescriptor2* @"\01??_R0H at 8" to i8*)) #1
+  %matches = icmp eq i32 %3, %4
+  br i1 %matches, label %catch, label %eh.resume
+
+catch:                                            ; preds = %lpad
+  %5 = extractvalue { i8*, i32 } %2, 0
+  %6 = tail call i8* @llvm.eh.begincatch(i8* %5) #1
+  %7 = bitcast i8* %6 to i32*
+  %8 = load i32* %7, align 4, !tbaa !7
+  %idxprom = sext i32 %NumExceptions.020 to i64
+  %arrayidx = getelementptr inbounds [10 x i32]* %ExceptionVal, i64 0, i64 %idxprom
+  store i32 %8, i32* %arrayidx, align 4, !tbaa !7
+  %inc = add nsw i32 %NumExceptions.020, 1
+  %cmp1 = icmp eq i32 %8, %i.019
+  br i1 %cmp1, label %if.then, label %if.else
+
+if.then:                                          ; preds = %catch
+  %9 = load i32* %b, align 4, !tbaa !8
+  %add2 = add nsw i32 %9, %i.019
+  store i32 %add2, i32* %b, align 4, !tbaa !8
+  br label %if.end
+
+if.else:                                          ; preds = %catch
+  %10 = load i32* %a, align 8, !tbaa !2
+  %add4 = add nsw i32 %10, %8
+  store i32 %add4, i32* %a, align 8, !tbaa !2
+  br label %if.end
+
+if.end:                                           ; preds = %if.else, %if.then
+  tail call void @llvm.eh.endcatch() #1
+  br label %try.cont
+
+; CHECK: try.cont:                                         ; preds = %if.end, %invoke.cont
+; CHECK-NOT:  %NumExceptions.1 = phi i32 [ %NumExceptions.020, %invoke.cont ], [ %inc, %if.end ]
+; CHECK:   %NumExceptions.1 = phi i32 [ %NumExceptions.020.reload, %invoke.cont ], [ %inc, %if.end ]
+; CHECK:   tail call void @"\01?does_not_throw@@YAXH at Z"(i32 %NumExceptions.1)
+; CHECK-NOT:  %inc5 = add nuw nsw i32 %i.019, 1
+; CHECK:   %inc5 = add nuw nsw i32 %i.019.reload, 1
+; CHECK:   %cmp = icmp slt i32 %inc5, 10
+; CHECK:   store i32 %NumExceptions.1, i32* %NumExceptions.020.reg2mem
+; CHECK:   store i32 %inc5, i32* %i.019.reg2mem
+; CHECK:   br i1 %cmp, label %for.body, label %for.end
+
+try.cont:                                         ; preds = %if.end, %invoke.cont
+  %NumExceptions.1 = phi i32 [ %NumExceptions.020, %invoke.cont ], [ %inc, %if.end ]
+  tail call void @"\01?does_not_throw@@YAXH at Z"(i32 %NumExceptions.1)
+  %inc5 = add nuw nsw i32 %i.019, 1
+  %cmp = icmp slt i32 %inc5, 10
+  br i1 %cmp, label %for.body, label %for.end
+
+for.end:                                          ; preds = %try.cont
+  %NumExceptions.1.lcssa = phi i32 [ %NumExceptions.1, %try.cont ]
+  %arraydecay = getelementptr inbounds [10 x i32]* %ExceptionVal, i64 0, i64 0
+  call void @"\01?dump@@YAXPEAHHAEAUSomeData@@@Z"(i32* %arraydecay, i32 %NumExceptions.1.lcssa, %struct.SomeData* dereferenceable(8) %tmpcast)
+  call void @llvm.lifetime.end(i64 40, i8* %0) #1
+  ret void
+
+eh.resume:                                        ; preds = %lpad
+  %.lcssa = phi { i8*, i32 } [ %2, %lpad ]
+  resume { i8*, i32 } %.lcssa
+}
+
+; The following catch handler should be outlined.
+; CHECK: define i8* @"\01?test@@YAXXZ.catch"(i8*, i8*) {
+; CHECK: catch.entry:
+; CHECK:   %eh.alloc = call i8* @llvm.framerecover(i8* bitcast (void ()* @"\01?test@@YAXXZ" to i8*), i8* %1)
+; CHECK:   %eh.data = bitcast i8* %eh.alloc to %"struct.\01?test@@YAXXZ.ehdata"*
+; CHECK:   %eh.obj.ptr = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 1
+; CHECK:   %eh.obj = load i8** %eh.obj.ptr
+; CHECK:   %eh.temp.alloca = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 2
+; CHECK:   %NumExceptions.020.reload = load i32* %eh.temp.alloca
+; CHECK:   %ExceptionVal = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 3
+; CHECK:   %eh.temp.alloca1 = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 4
+; CHECK:   %i.019.reload = load i32* %eh.temp.alloca1
+; CHECK:   %eh.temp.alloca2 = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 5
+; CHECK:   %a.reload = load i32** %eh.temp.alloca2
+; CHECK:   %eh.temp.alloca3 = getelementptr inbounds %"struct.\01?test@@YAXXZ.ehdata"* %eh.data, i32 0, i32 6
+; CHECK:   %b.reload = load i32** %eh.temp.alloca3
+; CHECK:   %2 = bitcast i8* %eh.obj to i32*
+; CHECK:   %3 = load i32* %2, align 4, !tbaa !7
+; CHECK:   %idxprom = sext i32 %NumExceptions.020.reload to i64
+; CHECK:   %arrayidx = getelementptr inbounds [10 x i32]* %ExceptionVal, i64 0, i64 %idxprom
+; CHECK:   store i32 %3, i32* %arrayidx, align 4, !tbaa !7
+; CHECK:   %inc = add nsw i32 %NumExceptions.020.reload, 1
+; CHECK:   %cmp1 = icmp eq i32 %3, %i.019.reload
+; CHECK:   br i1 %cmp1, label %if.then, label %if.else
+;
+; CHECK: if.then:                                          ; preds = %catch.entry
+; CHECK:   %4 = load i32* %b.reload, align 4, !tbaa !8
+; CHECK:   %add2 = add nsw i32 %4, %i.019.reload
+; CHECK:   store i32 %add2, i32* %b.reload, align 4, !tbaa !8
+; CHECK:   br label %if.end
+;
+; CHECK: if.else:                                          ; preds = %catch.entry
+; CHECK:   %5 = load i32* %a.reload, align 8, !tbaa !2
+; CHECK:   %add4 = add nsw i32 %5, %3
+; CHECK:   store i32 %add4, i32* %a.reload, align 8, !tbaa !2
+; CHECK:   br label %if.end
+;
+; CHECK: if.end:                                           ; preds = %if.else, %if.then
+; CHECK:   ret i8* blockaddress(@"\01?test@@YAXXZ", %try.cont)
+; CHECK: }
+
+; Function Attrs: nounwind
+declare void @llvm.lifetime.start(i64, i8* nocapture) #1
+
+declare void @"\01?may_throw@@YAXXZ"() #2
+
+declare i32 @__CxxFrameHandler3(...)
+
+; Function Attrs: nounwind readnone
+declare i32 @llvm.eh.typeid.for(i8*) #3
+
+declare i8* @llvm.eh.begincatch(i8*)
+
+declare void @llvm.eh.endcatch()
+
+declare void @"\01?does_not_throw@@YAXH at Z"(i32) #2
+
+declare void @"\01?dump@@YAXPEAHHAEAUSomeData@@@Z"(i32*, i32, %struct.SomeData* dereferenceable(8)) #2
+
+; Function Attrs: nounwind
+declare void @llvm.lifetime.end(i64, i8* nocapture) #1
+
+attributes #0 = { uwtable "less-precise-fpmad"="false" "no-frame-pointer-elim"="false" "no-infs-fp-math"="false" "no-nans-fp-math"="false" "stack-protector-buffer-size"="8" "unsafe-fp-math"="false" "use-soft-float"="false" }
+attributes #1 = { nounwind }
+attributes #2 = { "less-precise-fpmad"="false" "no-frame-pointer-elim"="false" "no-infs-fp-math"="false" "no-nans-fp-math"="false" "stack-protector-buffer-size"="8" "unsafe-fp-math"="false" "use-soft-float"="false" }
+attributes #3 = { nounwind readnone }
+
+!llvm.module.flags = !{!0}
+!llvm.ident = !{!1}
+
+!0 = !{i32 1, !"PIC Level", i32 2}
+!1 = !{!"clang version 3.7.0 (trunk 228868)"}
+!2 = !{!3, !4, i64 0}
+!3 = !{!"?AUSomeData@@", !4, i64 0, !4, i64 4}
+!4 = !{!"int", !5, i64 0}
+!5 = !{!"omnipotent char", !6, i64 0}
+!6 = !{!"Simple C/C++ TBAA"}
+!7 = !{!4, !4, i64 0}
+!8 = !{!3, !4, i64 4}





More information about the llvm-commits mailing list
--- llvm/trunk/lib/CodeGen/WinEHPrepare.cpp (original)
+++ llvm/trunk/lib/CodeGen/WinEHPrepare.cpp Mon Mar  2 18:41:03 2015
@@ -46,7 +46,7 @@ struct HandlerAllocas {
 // allocation block, and to remap the frame variable allocas (including
 // spill locations as needed) to GEPs that get the variable from the
 // frame allocation structure.
-typedef MapVector<AllocaInst *, HandlerAllocas> FrameVarInfoMap;
+typedef MapVector<Value *, HandlerAllocas> FrameVarInfoMap;
 
 class WinEHPrepare : public FunctionPass {
   std::unique_ptr<FunctionPass> DwarfPrepare;
@@ -261,22 +261,31 @@ bool WinEHPrepare::prepareCPPEHHandlers(
   // all the entries in the HandlerData have been processed this isn't a
   // problem.
   for (auto &VarInfoEntry : FrameVarInfo) {
-    AllocaInst *ParentAlloca = VarInfoEntry.first;
+    Value *ParentVal = VarInfoEntry.first;
     HandlerAllocas &AllocaInfo = VarInfoEntry.second;
 
-    // If the instruction still has uses in the parent function or if it is
-    // referenced by more than one handler, add it to the frame allocation
-    // structure.
-    if (ParentAlloca->getNumUses() != 0 || AllocaInfo.Allocas.size() > 1) {
-      Type *VarTy = ParentAlloca->getAllocatedType();
+    if (auto *ParentAlloca = dyn_cast<AllocaInst>(ParentVal)) {
+      // If the instruction still has uses in the parent function or if it is
+      // referenced by more than one handler, add it to the frame allocation
+      // structure.
+      if (ParentAlloca->getNumUses() != 0 || AllocaInfo.Allocas.size() > 1) {
+        Type *VarTy = ParentAlloca->getAllocatedType();
+        StructTys.push_back(VarTy);
+        AllocaInfo.ParentFrameAllocationIndex = Idx++;
+      } else {
+        // If the variable is not used in the parent frame and it is only used
+        // in one handler, the alloca can be removed from the parent frame
+        // and the handler will keep its "temporary" alloca to define the value.
+        // An element index of -1 is used to indicate this condition.
+        AllocaInfo.ParentFrameAllocationIndex = -1;
+      }
+    } else {
+      // FIXME: Sink non-alloca values into the handler if they have no other
+      //        uses in the parent function after outlining and are only used in
+      //        one handler.
+      Type *VarTy = ParentVal->getType();
       StructTys.push_back(VarTy);
       AllocaInfo.ParentFrameAllocationIndex = Idx++;
-    } else {
-      // If the variable is not used in the parent frame and it is only used
-      // in one handler, the alloca can be removed from the parent frame
-      // and the handler will keep its "temporary" alloca to define the value.
-      // An element index of -1 is used to indicate this condition.
-      AllocaInfo.ParentFrameAllocationIndex = -1;
     }
   }
 
@@ -331,11 +340,41 @@ bool WinEHPrepare::prepareCPPEHHandlers(
   // Finally, replace all of the temporary allocas for frame variables used in
   // the outlined handlers and the original frame allocas with GEP instructions
   // that get the equivalent pointer from the frame allocation struct.
+  Instruction *FrameEHDataInst = cast<Instruction>(FrameEHData);
+  BasicBlock::iterator II = FrameEHDataInst;
+  ++II;
+  Instruction *AllocaInsertPt = II;
   for (auto &VarInfoEntry : FrameVarInfo) {
-    AllocaInst *ParentAlloca = VarInfoEntry.first;
+    Value *ParentVal = VarInfoEntry.first;
     HandlerAllocas &AllocaInfo = VarInfoEntry.second;
     int Idx = AllocaInfo.ParentFrameAllocationIndex;
 
+    // If the mapped value isn't already an alloca, we need to spill it if it
+    // is a computed value or copy it if it is an argument.
+    AllocaInst *ParentAlloca = dyn_cast<AllocaInst>(ParentVal);
+    if (!ParentAlloca) {
+      if (auto *Arg = dyn_cast<Argument>(ParentVal)) {
+        // Lower this argument to a copy and then demote that to the stack.
+        // We can't just use the argument location because the handler needs
+        // it to be in the frame allocation block.
+        // Use 'select i8 true, %arg, undef' to simulate a 'no-op' instruction.
+        Value *TrueValue = ConstantInt::getTrue(Context);
+        Value *UndefValue = UndefValue::get(Arg->getType());
+        Instruction *SI =
+            SelectInst::Create(TrueValue, Arg, UndefValue,
+                               Arg->getName() + ".tmp", AllocaInsertPt);
+        Arg->replaceAllUsesWith(SI);
+        // Reset the select operand, because it was clobbered by the RAUW above.
+        SI->setOperand(1, Arg);
+        ParentAlloca = DemoteRegToStack(*SI, true, SI);
+      } else if (auto *PN = dyn_cast<PHINode>(ParentVal)) {
+        ParentAlloca = DemotePHIToStack(PN, AllocaInsertPt);
+      } else {
+        Instruction *ParentInst = cast<Instruction>(ParentVal);
+        ParentAlloca = DemoteRegToStack(*ParentInst, true, ParentInst);
+      }
+    }
+
     // If we have an index of -1 for this instruction, it means it isn't used
     // outside of this handler.  In that case, we just keep the "temporary"
     // alloca in the handler and erase the original alloca from the parent.
@@ -353,6 +392,8 @@ bool WinEHPrepare::prepareCPPEHHandlers(
       ParentAlloca->replaceAllUsesWith(ElementPtr);
       ParentAlloca->removeFromParent();
       ElementPtr->takeName(ParentAlloca);
+      if (ParentAlloca == AllocaInsertPt)
+        AllocaInsertPt = dyn_cast<Instruction>(ElementPtr);
       delete ParentAlloca;
 
       // Next replace all outlined allocas that are mapped to it.
@@ -589,38 +630,33 @@ WinEHFrameVariableMaterializer::WinEHFra
 }
 
 Value *WinEHFrameVariableMaterializer::materializeValueFor(Value *V) {
-  // If we're asked to materialize an alloca variable, we temporarily
-  // create a matching alloca in the outlined function.  When all the
-  // outlining is complete, we'll collect these into a structure and
-  // replace these temporary allocas with GEPs referencing the frame
-  // allocation block.
+  // If we're asked to materialize a value that is an instruction, we
+  // temporarily create an alloca in the outlined function and add this
+  // to the FrameVarInfo map.  When all the outlining is complete, we'll
+  // collect these into a structure, spilling non-alloca values in the
+  // parent frame as necessary, and replace these temporary allocas with
+  // GEPs referencing the frame allocation block.
+
+  // If the value is an alloca, the mapping is direct.
   if (auto *AV = dyn_cast<AllocaInst>(V)) {
-    AllocaInst *NewAlloca = Builder.CreateAlloca(
-        AV->getAllocatedType(), AV->getArraySize(), AV->getName());
+    AllocaInst *NewAlloca = dyn_cast<AllocaInst>(AV->clone());
+    Builder.Insert(NewAlloca, AV->getName());
     FrameVarInfo[AV].Allocas.push_back(NewAlloca);
     return NewAlloca;
   }
 
-// FIXME: Do PHI nodes need special handling?
-
-// FIXME: Are there other cases we can handle better?  GEP, ExtractValue, etc.
-
-// FIXME: This doesn't work during cloning because it finds an instruction
-//        in the use list that isn't yet part of a basic block.
-#if 0
-  // If we're asked to remap some other instruction, we'll need to
-  // spill it to an alloca variable in the parent function and add a
-  // temporary alloca in the outlined function to be processed as
-  // described above.
-  Instruction *Inst = dyn_cast<Instruction>(V);
-  if (Inst) {
-    AllocaInst *Spill = DemoteRegToStack(*Inst, true);
-    AllocaInst *NewAlloca = Builder.CreateAlloca(Spill->getAllocatedType(),
-                                                 Spill->getArraySize());
-    FrameVarMap[AV] = NewAlloca;
-    return NewAlloca;
+  // For other types of instructions or arguments, we need an alloca based on
+  // the value's type and a load of the alloca.  The alloca will be replaced
+  // by a GEP, but the load will stay.  In the parent function, the value will
+  // be spilled to a location in the frame allocation block.
+  if (isa<Instruction>(V) || isa<Argument>(V)) {
+    AllocaInst *NewAlloca =
+        Builder.CreateAlloca(V->getType(), nullptr, "eh.temp.alloca");
+    FrameVarInfo[V].Allocas.push_back(NewAlloca);
+    LoadInst *NewLoad = Builder.CreateLoad(NewAlloca, V->getName() + ".reload");
+    return NewLoad;
   }
-#endif
 
+  // Don't materialize other values.
   return nullptr;
 }

Added: llvm/trunk/test/CodeGen/X86/cppeh-inalloca.ll
URL: http://llvm.org/viewvc/llvm-project/llvm/trunk/test/CodeGen/X86/cppeh-inalloca.ll?rev=231042&view=auto
==============================================================================